// Code to build a graph based on an image
// Requires a graphical library, to handle/display images...
// You might just update this code to your own graphical library

#ifndef REQUIRES_GRAPHICAL_LIBRARY_H
#define REQUIRES_GRAPHICAL_LIBRARY_H

#include <Imagine/Graphics.h>
#include <Imagine/Images.h>
using namespace Imagine;

#include "images.h"
#include "with_graphical_library.h"

#include "approximation.h"

namespace MRF_F0 {


/*********** A few simple tools first ************/

void add_noise(ColorImage& A, double niv = 180){
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++){
      A(i,j).r() += (rand() * niv / RAND_MAX);
      A(i,j).g() += (rand() * niv / RAND_MAX);
      A(i,j).b() += (rand() * niv / RAND_MAX);
    }  
}


double norme(Color C){
  return sqrt(square(double(C.r())) + square( double(C.g()) ) + square( double(C.b()) ));
}



void fill(int x, vector<vector<int> >& V, int a, int b, int c, int d){
  V[x].resize(4);
  V[x][0] = a;
  V[x][1] = b;
  V[x][2] = c;
  V[x][3] = d;
}






/******************* Code to build a graph based on an image *******************/

// Two examples shown (see a few lines later to pick one)

// Distance : 0 = L2 norm, 1 = Kullback-Leibler

 void minimize_on_an_image_grid(int distance, bool verbose_basic = true, bool verbose_a_bit_more = true, bool verbose_a_lot = false, bool verbose_graph = false, bool test_stupide = true, bool test_local = false, bool test_global = false, ostream& cout = std::cout){

  if (verbose_basic){
    cout << TextGreen << "**********************************************************" << endl;
    cout << "***************" << TextRed << "  Creation of an energy   " << TextGreen << "*****************" << endl;
    cout << "**********************************************************" << TextNormal << endl;
  }

// Two examples are shown : one with an uploaded image, and one by building an image.
// Comment next line to build an image instead of loading one
//#define Load_an_image 

// In both cases, the code produces various images and save them at predefined locations (so, previous ones are replaced).
// Code starts with building the energy (two possibilities) and then common code again.

#ifdef Load_an_image

/*******************************************************************************************************/
/********************** Creating an energy based on an image loaded (Joconde example) ******************/
/*******************************************************************************************************/

  // Load an image
  ColorImage Joconde;
  Joconde.load("/user/gcharpia/home/prog/Segmentation_MRF/joconde.jpg", 1);
  //  add_noise(Joconde);

  int& width = Joconde.nx; // grid width
  int& length = Joconde.ny;

  int nb_var = width * length; // nb variables
  if (verbose_basic)
    cout << "Nb variables : " << nb_var << " on a grid " << width << " x " << length << endl;

  // nb of labels, here not depending on the variable
  int nb_labl = 8; 

  // To display in the end
  int zoom = 1; 

  /******** Creating potentials, matrices and permutations ********/
  // searching for extremal intensity values
  double int_min = norme(Joconde(0,0)); 
  double int_max = int_min;
  
  for (int i = 0; i<Joconde.nx; i++)
    for (int j = 0; j<Joconde.ny; j++){
      double a = norme(Joconde(i,j));
      if (a > int_max)
	int_max = a;
      if (a < int_min)
	int_min = a;
  }
  if (int_min == int_max)
    int_max += 1;

  // Potentials of each label for each variable : Pot[variable index][label index] = real value
  vector< vector<double> > Pot(nb_var); 
  for (int i = 0; i<nb_var; i++){ 
    //    Pot[i] = random_vector(nb_labl, 1); // prefer random potential ?
    Pot[i].resize(nb_labl);
    Pot[i][0] = 0; Pot[i][7] = 0; // labels 0 and 7 are better ignored
    int i_x = i%width; int i_y = i/width;
    Color C = Joconde(i_x, i_y);
    double a = (norme(C) - int_min)/(int_max - int_min) ; //sqrt(square(double(C.r())) + square(double(C.g())) + square(double(C.b()))) / 256.;
    for (int k = 1; k<7; k++){
      Pot[i][k] = 10*(fabs(a - (k-0.5) / 6.)) - 10; // This is the potential used in the paper for the Joconde example
    }

    if (verbose_a_lot)
      cout << "Pot x_" << i << " : " << Pot[i] << endl;
  }
    
  // Neighborhood and interactions between variables
  vector< vector< pair<int, int> > > Neighbors(nb_var); // list, for each variable, of the indices of its neighbors (with also the index of the associated interaction matrix), initialized with "no neighbor"
  //  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // type used below
  Energy::MIV Interact; // interaction matrices with indices of variables involved
  int nb_mat_inter = 0;


  if (verbose_a_bit_more)
    cout << endl << "Basic interaction matrices used :" << endl;
  // basic interaction matrices known to be in F0 (generated from coefficients alpha, here named AT)
  DoubleImage AT(3,3); AT.nul();
  AT(0,0) = 1;
  AT(1,1) = 1;
  AT(2,2) = 1;
  DoubleImage ATcoh = alphas_to_matrix(AT); // used for modeling spatial coherency
  if (verbose_a_bit_more)
    ATcoh.display_text();

  AT.nul();
  AT(0,1) = 1;
  AT(1,0) = 1;
  AT(1,2) = 1;
  AT(2,1) = 1;
  AT(0,2) = 1;
  AT(2,0) = 1;
  DoubleImage ATvar = alphas_to_matrix(AT); // used for modeling that spatial variations are prefered
  if (verbose_a_bit_more)
    ATvar.display_text();

  // not used but could have been
  AT.nul();
  AT(0,1) = 1;
  AT(1,0) = 1;
  AT(0,2) = 1;
  AT(2,0) = 1;
  DoubleImage ATb = alphas_to_matrix(AT);
  if (verbose_a_bit_more){
    cout << endl << "(not used but could have been :)" << endl;
    ATb.display_text();
  }

  AT.nul();
  AT(0,1) = 1;
  AT(1,0) = 1;
  ATb.empty(); ATb = alphas_to_matrix(AT);
  if (verbose_a_bit_more)
    ATb.display_text();

  ATb.empty(); AT.empty();


  // Building interaction graph
  // (this code needs to be rewritten : it creates random interaction matrices, and later we replace them by meaningful ones...)
  // (still, the interaction graph is correct)
  double ampl_interact = 1. ; // a multiplying coefficient (just forget it)

  for (int i = 0; i<width; i++) 
    for (int j = 0; j<length; j++){
      int num_pix = i + j * width;

      // Add interaction with next pixel vertically
      if (i != (width-1)){
	int num2 = i+1 + j*width; // > num_pix
	DoubleImage M = random_matrix(nb_labl, nb_labl, ampl_interact); // Creates a random matrix, put yours instead 

	Interact.push_back( make_pair(M, make_pair(num_pix,num2)) ); // be careful : indices of variables have to be in increasing order, ie i<j => M = Mij and not Mji   // this is *IMPORTANT*
	Neighbors[num_pix].push_back( make_pair(num2, nb_mat_inter) );
	Neighbors[num2].push_back( make_pair(num_pix, nb_mat_inter) );
	nb_mat_inter++;
      }
  
      // Add interaction with next pixel horizontally
      if (j != (length-1)){
	int num2 = i + (j+1)*width; // > num_pix
	DoubleImage M = random_matrix(nb_labl, nb_labl, ampl_interact); // Creates a random matrix, put yours instead

	Interact.push_back( make_pair(M, make_pair(num_pix,num2)) ); // be careful : indices of variables have to be in increasing order, ie i<j => M = Mij and not Mji   // this is *IMPORTANT*
	Neighbors[num_pix].push_back( make_pair(num2, nb_mat_inter) );
	Neighbors[num2].push_back( make_pair(num_pix, nb_mat_inter) );
	nb_mat_inter++;
      }
    }

  if (verbose_basic)
    cout << "Nb interactions : " << nb_mat_inter << endl;



  // Color image to display potentials
  ColorImage PicPot(width, length); 

  for (int pix = 0; pix < nb_var; pix++ ){
    byte b = byte(255. * find_smallest(Pot[pix]) / 8.);
    PicPot(pix%width, pix/width) = Color(b,b,b);
  }
  PicPot.save("rectanglePotInit.png");


  // Building an edge detector to impose label variations there (this does not make sense, rather just put the edges you want, even by hand... it's just an example !)
  DoubleImage Joc2(Joconde.nx, Joconde.ny);
  for (int i = 0; i<Joc2.nx; i++)
    for (int j = 0; j<Joc2.ny; j++)
      Joc2(i,j) = norme(Joconde(i,j));
  
  smooth(Joc2,2,1);  // Joc2 is a greyscale, slightly smoothed version of the Joconde


  // To store edge lcoation
  RecImage<bool> Masque_x(width, length);
  RecImage<bool> Masque_y(width, length);
  for (int i = 0; i<width; i++)
    for (int j = 0; j<length; j++){
      Masque_x(i,j) = false;
      Masque_y(i,j) = false;
    }

  // Going through all interaction matrices and replacing them with something reasonable now
  for (int m = 0; m<int(Interact.size()); m++){
    DoubleImage& M = Interact[m].first;

    int pi = Interact[m].second.first;
    int pj = Interact[m].second.second;
    int pi_x = pi%width; int pi_y = pi/width;
    int pj_x = pj%width; int pj_y = pj/width;

    double a = fabs(Joc2(pi_x, pi_y) - Joc2(pj_x, pj_y))/256.;
    if (a > .09){
      //      copy(A2b, M);  // that was in the 4 label case
      copy(ATvar, M);  // 8 label case : enhancing label variations
      for (int i = 0; i<M.nx; i++)
	for (int j = 0; j<M.ny; j++)
	  M(i,j) *= 0.15 * a;
      Masque_x(pi_x, pi_y) = true; // just to display where we detected edges
    } else {
      //      copy(A1b, M);
      copy(ATcoh, M); // enhancing spatial coherency
      for (int i = 0; i<M.nx; i++)
	for (int j = 0; j<M.ny; j++)
	  M(i,j) *= 0.137/(a+0.001);
      Masque_x(pi_x,pi_y) = false;
    }
  }



#else
/******************************************************************************************/
/******************* Create your own image and energy (toy examples) **********************/
/******************************************************************************************/

  int width = 20;
  int length = width;

  // To display in the end
  int zoom = 5; 

  int nb_var = width * length; // nb variables
  if (verbose_basic)
    cout << "Nb variables : " << nb_var << " on a grid " << width << " x " << length << endl;

  // nb of labels, here not depending on the variable
  int nb_labl = 4; 

  // Display variables
  bool verbose_multi = false;
  bool verbose_coherence = false;
  bool verbose_coherence_pot = false;
  bool verbose_global = false;

  /******** Creating potentials, matrices and permutations ********/
  // Potentials of each label for each variable : Pot[variable index][label index] = real value
  vector< vector<double> > Pot(nb_var); 

  // Neighborhood and interactions between variables
  vector< vector< pair<int, int> > > Neighbors(nb_var); // list, for each variable, of the indices of its neighbors (with also the index of the associated interaction matrix), initialized with "no neighbor"
  //  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // type used below
  Energy::MIV Interact; // interaction matrices with indices of variables involved
  int nb_mat_inter = 0;


  // basic interaction matrices known to be in F0 (generated by coefficients alphas)
  DoubleImage A1(2,2); A1.nul();  // A1 : coherency
  A1(0,0) = 1; A1(1,1) = 1; A1(0,1) = 0; A1(1,0) = 0;

  DoubleImage A2(2,2); A2.nul(); // A2 : transition blue/red
  A2(1,0) = 1; A2(0,1) = 1; A2(0,0) = 0; A2(1,1) = 0;

  DoubleImage A1b = alphas_to_matrix(A1);
  if (verbose_a_bit_more){
    cout << endl << "Basic interaction matrices used :" << endl;
    A1b.display_text();
  }
  for (int i = 0; i<A1b.nx; i++)
    for (int j = 0; j<A1b.ny; j++)
      A1b(i,j) *= 1;

  DoubleImage A2b = alphas_to_matrix(A2);
  if (verbose_a_bit_more)
    A2b.display_text();

  // To give more/less weight to the second matrix (transition)
  //  for (int i = 0; i<A2b.nx; i++)
  //    for (int j = 0; j<A2b.ny; j++)
  //      A2b(i,j) *= 1;   // choose your weight


  // You can build directly the edges desired by settings masks (ie telling where are edges)
  RecImage<bool> Masque_x(width, length);
  RecImage<bool> Masque_y(width, length);
  for (int i = 0; i<width; i++)
    for (int j = 0; j<length; j++){  
      Masque_x(i,j) = false;
      Masque_y(i,j) = false;  // masks are empty now (no edge)
    }

  for (int i = width/4; i<3*width/4; i++){
    Masque_x(i,length/4) = true;
    Masque_x(i,length/2) = true;
    Masque_x(i,3*length/4) = true;  // 3 edges
  }


  // Idem : one can set the potentials directly and make sure they are in F0

  for (int i = 0; i<nb_var; i++){
    Pot[i].resize(nb_labl);

    int i_x = i%width; int i_y = i/width;

    if (  (i_x<3) ){ // I have a label preference in that region
      Pot[i][0] = 0;
      Pot[i][1] = -100;
      Pot[i][2] = 100;
      Pot[i][3] = 0;  // this vector is indeed in F0
    }
    else
      if (   ( (i_x<2*width/3)&&(i_x>width/3)&&(i_y == 2*length/3) ) || (i_x>width-4)  ){ // the opposite preference in another region
	Pot[i][0] = 0; 
	Pot[i][1] = 100;
	Pot[i][2] = -100;
	Pot[i][3] = 0;
      }
      else // neutral case
	{
	  Pot[i][0] = 0;
	  Pot[i][1] = 0;
	  Pot[i][2] = 0;
	  Pot[i][3] = 0;
	}
  }
  

  // Building interaction graph
  // (this code needs to be rewritten : it creates random interaction matrices, and later we replace them by meaningful ones...)
  // (still, the interaction graph is correct)
  double ampl_interact = 1. ; // a multiplying coefficient (just forget it)

  for (int i = 0; i<width; i++) 
    for (int j = 0; j<length; j++){
      int num_pix = i + j * width;

      // Add interaction with next pixel vertically
      if (i != (width-1)){
	int num2 = i+1 + j*width; // > num_pix
	DoubleImage M = random_matrix(nb_labl, nb_labl, ampl_interact); // Creates a random matrix, put yours instead 

	Interact.push_back( make_pair(M, make_pair(num_pix,num2)) ); // be careful : indices of variables have to be in increasing order, ie i<j => M = Mij and not Mji   // this is *IMPORTANT*
	Neighbors[num_pix].push_back( make_pair(num2, nb_mat_inter) );
	Neighbors[num2].push_back( make_pair(num_pix, nb_mat_inter) );
	nb_mat_inter++;
      }
  
      // Add interaction with next pixel horizontally
      if (j != (length-1)){
	int num2 = i + (j+1)*width; // > num_pix
	DoubleImage M = random_matrix(nb_labl, nb_labl, ampl_interact); // Creates a random matrix, put yours instead

	Interact.push_back( make_pair(M, make_pair(num_pix,num2)) ); // be careful : indices of variables have to be in increasing order, ie i<j => M = Mij and not Mji   // this is *IMPORTANT*
	Neighbors[num_pix].push_back( make_pair(num2, nb_mat_inter) );
	Neighbors[num2].push_back( make_pair(num_pix, nb_mat_inter) );
	nb_mat_inter++;
      }
    }

  if (verbose_basic)
    cout << "Nb interactions : " << nb_mat_inter << endl;


  // Color image to display potentials
  ColorImage PicPot(width, length); 

  for (int pix = 0; pix < nb_var; pix++ ){
    switch (find_smallest(Pot[pix])) {
    case 0 :
      PicPot(pix%width, pix/width) = BLACK; 
      break;
    case 1 :
      PicPot(pix%width, pix/width) = BLUE;
      break;
    case 2 :
      PicPot(pix%width, pix/width) = RED;
      break;
    case 3 :
      PicPot(pix%width, pix/width) = BLACK;
      break;
    }
  }
  PicPot.save("rectanglepot.png");


  // Going through all interaction matrices and replacing them with something reasonable now
  for (int m = 0; m<int(Interact.size()); m++){
    DoubleImage& M = Interact[m].first;

    int pi = Interact[m].second.first;
    int pj = Interact[m].second.second;
    int pi_x = pi%width; int pi_y = pi/width;
    int pj_x = pj%width; int pj_y = pj/width;

    // enforcing spatial coherence or label variation, depending on edge mask and relative location with respect to it.
   if (  ( Masque_x(pi_x,pi_y) && (pi_y < pj_y) )
    	  || ( Masque_x(pj_x,pj_y) && (pj_y < pi_y) )
    	  || ( Masque_y(pi_x,pi_y) && (pi_x < pj_x) )
    	  || ( Masque_y(pj_x,pj_y) && (pj_x < pi_x) ) ){
     copy(A2b, M); // transition zone
    } else {
     copy(A1b, M); // coherency zone
    }

   // adding noise to test stability of projection back to F0
   /*** for (int i = 0; i<M.nx; i++)
         for (int j = 0; j<M.ny; j++)
          M(i,j) += 1.001* (rand() * 1. / RAND_MAX);
  	   //       	M(i,j) = 0.1 * (fabs(i-j + (rand()%3)) + 0.001 * (rand() * 1. / RAND_MAX));
   ***/
  }

  // get rid of annoying compiler messages about unused variables
  if (verbose_a_lot && test_stupide && test_global){
    if (! verbose_a_lot)
      cout << distance << endl; // never reached
  }

#endif


  // Color Image to display edges chosen
  ColorImage PicEdge(width, length); 
  for (int i = 0; i<Masque_x.nx; i++)
    for (int j = 0; j<Masque_x.ny; j++)
      if (Masque_x(i,j) || Masque_y(i,j) )
  	PicEdge(i,j) = BLACK;
      else 
  	PicEdge(i,j) = Color(220,220,220);

  PicEdge.save("rectangleEdge.png");

  /******************************************************************************************/
  /******************** Now the energy is ready, back to main code ! ************************/
  /******************************************************************************************/

  Energy E(Pot, Neighbors, Interact, 0); 
  // Now the energy is ready.

  

  /*************************   canonization  *****************************/

  if (verbose_basic)
    cout << "Canonization" << endl;

  canonise(E);

  // You might want to do some nasty tricks when building your energy
  /*****  
     std::cout << endl << BoldBlue << " ********** Warning, remember this code changed the initial energy !!! ************ " << TextNormal << endl;
     double amplitude_potential = 1.;
     cout << "Potential amplitude after canonization : " << amplitude_potential << endl;
     for (int i = 0; i<E.nb_var; i++)
       for (int j = 0; j<int(E.Pot[i].size()); j++)
 	 E.Pot[i][j] *= amplitude_potential;
  *****/

  /************* The energy is ready, canonized ***********/
  // and it is named E

  // But you might have forgotten that the energy E you built has to belong to F0 !

  /*****************************************************************************************************/
  /************************** recognition, graph and minimization   ************************************/
  /*****************************************************************************************************/


  if (verbose_basic){ cout << endl;
    cout << TextGreen << "**********************************************************" << endl;
    cout << "*******" << TextRed << "   Creating the graph and minimization    " << TextGreen << "*********" << endl;
    cout << "**********************************************************" << TextNormal << endl;
  }
  
#ifdef Load_an_image
  //  Energy ApproxE; ApproxE.copy(E);
  PreGraph PG;
  if (verbose_basic)
    switch (distance) {
    case 0: 
      cout << "Distance L2 used for projection on F0." << endl; break;
    case 1:
      cout << "Kullback-Leibler divergence used for projection on F0." << endl; break;
    default :
      cout << "Oh, come on..." << endl;
    }
  construct_approximation(PG, E, distance, verbose_basic);
  vector<int>  Resultat = minimize_familial_MRF(PG, E,  test_stupide, test_local, test_global, verbose_basic, verbose_graph, verbose_a_lot);
#else
  // Honestly, this is silly, since we already know the coefficients alpha, but here try to find them as if we did not.
  // Should rather store the coefficients and use them directly,
  // the code would be almost instantaneous instead of taking ages (because of trying all possible representations whereas many many are locally correct).
  // I did write the fast, correct code once, but I don't find it. Anyway it's just a few lines.

  PreGraphMulti PG;
  construct_pregraph(PG, E, verbose_basic, verbose_a_bit_more, verbose_multi, verbose_coherence, verbose_coherence_pot, verbose_global, cout);
  vector<int>  Resultat = minimize_familial_MRF(E, PG, test_local, verbose_basic, verbose_graph);
#endif

  if (verbose_basic)
    cout << "Energy at solution found : " << evaluate_energy(E,Resultat) << endl ;
   
  
  // To display results

  if (verbose_basic)
    cout << "Displaying solution..." << endl;
  ColorImage PicLabel(width, length); 
  for (int pix = 0; pix < nb_var; pix++ ){

    if (nb_labl == 8){
      byte b = byte(255. * Resultat[pix] / 8.);
      PicLabel(pix%width, pix/width) = Color(b,b,b);
    }

    if (nb_labl == 4){
      //      cout << find_smallest(E.Pot[pix]) << "-";
      switch (Resultat[pix]) {
      case 0 :
	PicLabel(pix%width, pix/width) = WHITE;
	break;
      case 1 :
	PicLabel(pix%width, pix/width) = BLUE;
	break;
      case 2 :
	PicLabel(pix%width, pix/width) = RED;
	break;
      case 3 :
	PicLabel(pix%width, pix/width) = BLACK;
	break;
      }

      if (verbose_a_bit_more){
	cout << Resultat[pix] << " ";
	if (( (pix+1) % width) == 0)
	  cout << endl;
      }
    }
  }
  PicLabel.save("rectangleLabel.png");
  cout << endl;


  // Display in windows
  Window W0 = openWindow(zoom*PicPot.nx, zoom*PicPot.ny);
  setActiveWindow(W0);
  PicPot.display(zoom);

  Window WA = openWindow(zoom*PicEdge.nx, zoom*PicEdge.ny);
  setActiveWindow(WA);
  PicEdge.display(zoom);

  Window WB = openWindow(zoom*PicLabel.nx, zoom*PicLabel.ny); 
  setActiveWindow(WB);
  PicLabel.display(zoom); 


  // Cleaning
  PG.empty();
  E.empty();
  Resultat.clear();
 
  endGraphics();

}


} // end of namespace

#endif
